import assert from "assert";
import { Chain, PublicClient, Transport, encodeFunctionData } from "viem";
import { Erc20, Erc4337, SmartAccountV1, UniswapV3 } from "./abi";
import { Dex, DexContractsV3, DexName } from "./types/dex";
import { UserOperation } from "./types/erc4337";
import { CreateSwapRequestInput, CreateSwapRequestOutput } from "./types/smartAccount";
import { DEX_CHAINS, ERC4337_CHAINS } from "./contracts";

export class SmartAccountV1Provider {
  publicClient: PublicClient<Transport, Chain>;
  dex?: Dex;
  entrypoint: `0x${string}`;
  smartAccountFactory: `0x${string}`;

  constructor(
    publicClient: PublicClient<Transport, Chain>,
    entrypoint?: `0x${string}`,
    smartAccountFactory?: `0x${string}`,
    dex?: { [key in DexName]?: DexContractsV3 },
  ) {
    this.publicClient = publicClient;

    if (dex) {
      this.dex = dex;
    } else {
      const dexChain = DEX_CHAINS[publicClient.chain.id];
      if (dexChain) {
        this.dex = dexChain;
      }
    }

    const erc4337Chain = ERC4337_CHAINS[publicClient.chain.id];
    if (entrypoint) {
      this.entrypoint = entrypoint;
    } else {
      if (!erc4337Chain || !erc4337Chain.entrypoint) {
        throw new Error("Entrypoint not supported on this chain");
      }
      this.entrypoint = erc4337Chain.entrypoint;
    }

    if (smartAccountFactory) {
      this.smartAccountFactory = smartAccountFactory;
    } else {
      if (!erc4337Chain || !erc4337Chain.smartAccountFactoryV1) {
        throw new Error("Smart account V1 not supported on this chain");
      }
      this.smartAccountFactory = erc4337Chain.smartAccountFactoryV1;
    }
  }

  async createSwapRequest(createSwapRequestInput: CreateSwapRequestInput): Promise<CreateSwapRequestOutput> {
    if (!createSwapRequestInput.smartAccount && !createSwapRequestInput.initSmartAccountInput) {
      throw new Error("Must provide either `smartAccount` or `initSmartAccountInput`");
    }

    if (!this.dex) {
      throw new Error("Dex not supported");
    }

    const dexName = createSwapRequestInput.dex;
    const dex = this.dex[dexName];
    if (!dex) {
      throw new Error(`${dexName} not supported`);
    }

    let smartAccount = createSwapRequestInput.smartAccount;
    let initCode = "0x" as `0x${string}`;

    const initSmartAccountInput = createSwapRequestInput.initSmartAccountInput;
    if (initSmartAccountInput) {
      const { owner, salt } = initSmartAccountInput;
      const address = await this.publicClient.readContract({
        abi: SmartAccountV1.FACTORY_ABI,
        address: this.smartAccountFactory,
        functionName: "getAddress",
        args: [owner, salt],
      });
      const bytecode = await this.publicClient.getBytecode({ address });
      if (bytecode && bytecode !== "0x") {
        throw new Error("Smart account already exists");
      }

      const initCodeCallData = encodeFunctionData({
        abi: SmartAccountV1.FACTORY_ABI,
        functionName: "createAccount",
        args: [owner, salt],
      });

      initCode = (this.smartAccountFactory.toLowerCase() + initCodeCallData.slice(2)) as `0x${string}`;
      smartAccount = address.toLowerCase() as `0x${string}`;
    }
    assert(smartAccount !== undefined);

    const { tokenIn, tokenOut, fee, recipient, deadline, amountIn, amountOutMinimum } =
      createSwapRequestInput.swapInput;
    const approveCallData = encodeFunctionData({
      abi: Erc20.ABI,
      functionName: "approve",
      args: [dex.swapRouter, amountIn],
    });
    const exactInputSingleCallData = encodeFunctionData({
      abi: UniswapV3.SWAP_ROUTER_ABI,
      functionName: "exactInputSingle",
      args: [
        {
          tokenIn,
          tokenOut,
          fee,
          recipient: recipient || smartAccount,
          deadline,
          amountIn,
          amountOutMinimum,
          sqrtPriceLimitX96: BigInt(0),
        },
      ],
    });

    const dest = [tokenIn, dex.swapRouter];
    const value: bigint[] = [];
    const func = [approveCallData, exactInputSingleCallData];

    const executeBatchCallData = encodeFunctionData({
      abi: SmartAccountV1.ABI,
      functionName: "executeBatch",
      args: [dest, value, func],
    });

    // nonce is uint256 with 192-bit key and 64-bit value
    // key is pool address
    const nonce = await this.getNonceForSmartAccountV1(smartAccount, dex.factory, [tokenIn, tokenOut], fee);

    // ==================================================
    // ===== maximum of values before is "< uin120" =====
    // ==================================================
    // Note: Step 1 -> 3 is gasLimit, 4 is gasPrice
    // So, (1 + 2 + 3) * 4 = total `wei` needs to execute this transaction

    // 1. calculate callGasLimit => Gas limit for execution phase, estimate by using EntryPoint simulation (function simulateValidation)
    const callGasLimit = await this.calculateExecuteBatchGasLimit(smartAccount, dest, value, func);

    // 2. calculate verificationGasLimit => Gas limit for verification phase, estimate by using EntryPoint simulation (function simulateValidation)
    const verificationGasLimit = await this.calculateVerificationGasLimitV1(initCode !== "0x");

    // 3. calculate preVerificationGas => Gas to compensate the bundler, Use internal bundler => Don't care
    const preVerificationGas = BigInt(0);

    // 4. calculate maxFeePerGas => 30 gwei
    // 5. calculate maxPriorityFeePerGas => estimate: 2 gwei
    // TODO: Both fees based on particular chain

    let maxFeePerGas = BigInt(0);
    let maxPriorityFeePerGas = BigInt(0);

    if (!createSwapRequestInput.gasless) {
      maxFeePerGas = BigInt(30e9);
      maxPriorityFeePerGas = BigInt(2e9);
    }

    const userOpWithoutSign: UserOperation = {
      sender: smartAccount,
      nonce,
      initCode,
      callData: executeBatchCallData,
      callGasLimit,
      verificationGasLimit,
      preVerificationGas,
      maxFeePerGas,
      maxPriorityFeePerGas,
      paymasterAndData: "0x" as `0x${string}`,
      signature: "0x" as `0x${string}`,
    };

    // userOpHash does not depend on signature
    const userOpHash = await this.publicClient.readContract({
      abi: Erc4337.ENTRYPOINT_ABI,
      address: this.entrypoint,
      functionName: "getUserOpHash",
      args: [userOpWithoutSign],
    });

    return { smartAccount, userOpHash, request: userOpWithoutSign };
  }

  async getNonceForSmartAccountV1(
    smartAccount: `0x${string}`,
    dexFactory: `0x${string}`,
    tokens: `0x${string}`[],
    fee: number,
  ): Promise<bigint> {
    if (tokens.length !== 2) {
      throw new Error("Invalid token pair");
    }

    const poolAddr = await this.publicClient.readContract({
      abi: UniswapV3.FACTORY_ABI,
      address: dexFactory,
      functionName: "getPool",
      args: [tokens[0], tokens[1], fee],
    });
    const formattedPoolAddr = poolAddr.toLowerCase();
    const key = BigInt(formattedPoolAddr);

    const nonce = await this.publicClient.readContract({
      abi: Erc4337.ENTRYPOINT_ABI,
      address: this.entrypoint,
      functionName: "getNonce",
      args: [smartAccount, key],
    });

    return nonce;
  }

  async calculateVerificationGasLimitV1(initSmartAccount?: boolean): Promise<bigint> {
    if (initSmartAccount) {
      return BigInt(300_000);
    } else {
      return BigInt(50_000);
    }
  }

  async calculateExecuteBatchGasLimit(
    smartAccount: `0x${string}`,
    dest: `0x${string}`[],
    value: bigint[],
    func: `0x${string}`[],
  ): Promise<bigint> {
    const estimatedGas = await this.publicClient.estimateContractGas({
      abi: SmartAccountV1.ABI,
      address: smartAccount,
      functionName: "executeBatch",
      args: [dest, value, func],
      account: this.entrypoint,
    });

    // ref: https://github.com/wolflo/evm-opcodes/blob/main/gas.md#aa-call-operations
    const baseGas = BigInt(21_000);

    // TODO: smart account does not exist, so can not estimate gas
    if (estimatedGas < 30_000) {
      return 170_000n;
    }

    return estimatedGas - baseGas;
  }
}
